local T, scan_sector, check_number = ...

local handler = {
	["commands"] = function(data)
		print("\127FFFFFF"..T.help["commands"])
	end,
	
	["pos"] = function(data)
		if (T.hasmdi) then return end

		local x, y = data[1], data[2]
		local container = T.UI.display_container
		if (check_number(x) and check_number(y)) then
			if (x:match("^[+-]")) then x = container.cx + x end
			if (y:match("^[+-]")) then y = container.cy + y end
			T.x = x
			T.y = y
			T:SaveSettings()
			T:RebuildDisplay()

		else
			if (data[1] == "reset") then
				T.x = nil
				T.y = nil
				print("\127FFFFFFTargetlist position has been reset to default.")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				local pos = container.cx.."x"..container.cy -- use the container positions since nil can't be concatenated
				if (not T.x) then pos = "unset (actual "..pos..")" end
				print("\127FFFFFFTargetlist position is "..pos)
			end
		end
	end,
	
	["fontscale"] = function(data)
		local newfontscale = check_number(data[1], T.defaults.fontscale.Lval, T.defaults.fontscale.Uval, T.defaults.fontscale.dp)
		if (newfontscale) then
			newfontscale = newfontscale < 0.1 and 0.1 or newfontscale
			T.fontscale = newfontscale
			print("\127FFFFFFFont size has been changed to "..newfontscale..".")
			T:SaveSettings()
			T:RebuildDisplay()
		else
			if (data[1] == "reset") then
				T.fontscale = T.defaults.fontscale.val
				print("\127FFFFFFThe font size has been been reset to "..T.fontscale..".")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				print("\127FFFFFFThe current font size is "..T.fontscale..".")
			end
		end
	end,
	
	["help"] = function(data)
		local word = table.remove(data, 1)
		if (word) then
			word = string.lower(tostring(word))
			if (type(T.help[word]) == "string") then
				print("\127FFFFFF"..T.help[word])
			end

		else
			print("\127FFFFFF"..T.help["help"])
		end
	end,
              
	["highlight"] = function(data)
		local r = check_number(data[1], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local g = check_number(data[2], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local b = check_number(data[3], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		local a = check_number(data[4], T.defaults.target_color.Lval, T.defaults.target_color.Uval)
		if (r and g and b) then
			a = a or 65
			T.current_target_color = string.format("%d %d %d %d *", r, g, b, a)
			print("\127FFFFFFThe highlight color has been changed to "..T.current_target_color:match("(.*) %*"))
			T:SaveSettings()
		else
			if (data[1] == "reset") then
				T.current_target_color = T.defaults.target_color.val
				print("\127FFFFFFThe highlight color has been reset to "..T.current_target_color:match("(.*) %*"))
				T:SaveSettings()
			else
				print("\127FFFFFFThe current highlight color is "..T.current_target_color:match("(.*) %*"))
			end
		end
	end,
	
	["rows"] = function(data)
		local newrows = check_number(data[1], T.defaults.list_count.Lval)
		if (newrows) then
			T.list_count = newrows
			print("\127FFFFFFThe number of rows has been changed to "..newrows..".")
			T:SaveSettings()
			T:RebuildDisplay()
		else
			if (data[1] == "reset") then
				T.list_count = T.defaults.list_count.val
				print("\127FFFFFFThe number of rows has been been reset to "..T.list_count)
				T:SaveSettings()
				T:RebuildDisplay()
			else
				print("\127FFFFFFThe current number of rows is "..T.list_count)
			end
		end
	end,

	["width"] = function(data)
		local newwidth = check_number(data[1], T.defaults.name_width.Lval)
		if (newwidth) then
			T.name_width = newwidth
			print("\127FFFFFFName width has been changed to "..newwidth)
			T:SaveSettings()
			T:RebuildDisplay()

		else
			if (data[1] == "reset") then
				T.name_width = nil
				print("\127FFFFFFName width has been reset.")
				T:SaveSettings()
				T:RebuildDisplay()
			else
				local width = T.display[1][2][3][2].w
				if (not T.name_width) then width = "unset (actual "..width..")" end
				print("\127FFFFFFThe current name width is "..width..".")
			end
		end
	end,

	["options"] = function()
		ShowDialog(T.optionsdlg)
	end,
}

local function cmd_processor(_, data)
	if (not data) then
		handler["options"]()
-- 		print("\127FFFFFF"..T.help["help"])
		return
	end
	local cmd = string.lower(table.remove(data, 1))
	if (type(handler[cmd]) == "function") then handler[cmd](data) end
end

local function exec_tab_func(func, ...)
	if (type(func) == "function") then
		local success, err = pcall(func, ...)
		if (not success) then
			console_print("ERROR: "..err)
			targetlist.API.UpdateDisplay(..., {{"ERROR", "ERROR", "ERROR"}})
		end
	end
end

function T:RefreshTab()
	local tab_tbl = T.tab_identifiers[T.tab_index]
	ProcessEvent("TARGETLIST_TAB_CHANGED") -- Don't pass the identifier.  This could allow other plugins to interfere.
	exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, T.filter, T.faction_filter) -- execute after the event in case the event handlers need to do some intialization
end


local function change_tab(i)
	local window = T.tab_window_size
	local offset = T.tab_window_offset
	local num_tabs = #T.tab_identifiers
	i = i < 0 and 0 or i > num_tabs and num_tabs or i
	local adj = i - T.tab_index

	if     (i == 1 and adj < i) then -- wrapping from right around to left
		offset = 0
	elseif (i == 0) then -- wrapping from left around to right
		i = num_tabs
		if (num_tabs > window) then
			offset = num_tabs - window
		end
	else
		-- Moving from right-to-left
		if (i < T.tab_index) then
			offset = i - 1
			if (offset > 0) then
				offset = offset - 1
			end

		-- Moving from left-to-right
		elseif (i > T.tab_index) then
			offset = i - window + 1
		end

		-- don't move the window off the left end
		if (offset >= num_tabs - window) then offset = num_tabs - window end
		-- don't move the window off the left end
		if (offset < 0) then offset = 0 end

	end
	T.tab_window_offset = offset
	T.tab_index = i

	T.filter, T.faction_filter = T.tab_filter_positions:GetPositions(T.tab_index) -- legacy variable.  set to avoid rewriting core loops
	T:RebuildDisplay()

	if (T.tab_index > 1) then
		T.refresh_timer:Kill()
		UnregisterEvent(T.Events, "TARGET_CHANGED")
	else
		scan_sector()
		RegisterEvent(T.Events, "TARGET_CHANGED")
	end
end


local function tab_next(_, data)
	local num = #T.tab_identifiers
	if (num == 1) then return end

	if (data) then
		data = tonumber(data[1]) % num
		if (data == 0) then data = num end
	end

	change_tab(data or ((T.tab_index % num) + 1))
end


local function tab_prev()
	local num = #T.tab_identifiers
	if (num == 1) then return end
	change_tab(((T.tab_index - 1) % num))
end

local function cycle_filter(_, data)
	local findex, fpage = T.tab_filter_positions:GetPositions(T.tab_index)
	if (not findex) then return end -- no filters to cycle

	local tab_tbl = T.tab_identifiers[T.tab_index]
	local filters = tab_tbl["filters"]
	local cur_filter = T.tab_filter_positions[tab_tbl]

	T.UI.filters[findex]:SetActive(false)

	local r, num_filters = pcall(function() return #filters end)

	local max_filters = (r and num_filters or 1)
	if (data) then
		data = tonumber(data[1]) % max_filters
		if (data == 0) then data = max_filters end
	end

	findex = data or ((findex) % max_filters) + 1
	T.filter = T.tab_filter_positions:SetPositions(T.tab_index, findex, nil) -- legacy variable.  set to avoid rewriting core loops
	findex, fpage = T.tab_filter_positions:GetPositions(T.tab_index) -- findex changed,  fpage is now inaccurate.  get it again

	if (filters["link_filter_positions"]) then
		T.UI.filters[findex]:SetBGColor(filters[1]["pages"][fpage]["bgcolor"], tab_tbl["colorsource"])
	end
	T.UI.filters[findex]:SetActive(true)

	if (T.tab_index > 1) then
		exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, findex, fpage)
	end

end

local function cycle_filter_page()
	local findex, fpage = T.tab_filter_positions:GetPositions(T.tab_index)
	local tab_tbl = T.tab_identifiers[T.tab_index]
	local filters = tab_tbl["filters"] --tab_tbl["filters"][findex]
	local filter = filters and filters[findex]
	local pages = filter and filter["pages"]
	local link = filters and filters["link_filter_positions"]

	if (not pages) then return end -- If there are no pages for this filter, then we're done

	local r, num_filter_pages = pcall(function() return #pages end)
	fpage = ((fpage) % (r and num_filter_pages or 1)) + 1
	T.faction_filter = T.tab_filter_positions:SetPositions(T.tab_index, nil, fpage) -- legacy variable.  set to avoid rewriting core loops

	local title
	local fgcolor
	local bgcolor
	title   = link and filter["title"] or pages[fpage]["title"]
	fgcolor = link and filters[1]["pages"][fpage]["fgcolor"] or pages[fpage]["fgcolor"] or filter["fgcolor"]
	bgcolor = link and filters[1]["pages"][fpage]["bgcolor"] or pages[fpage]["bgcolor"] or filter["bgcolor"]

	title = title or filter["title"]
	fgcolor = fgcolor
	bgcolor = bgcolor

	T.UI.filters[findex]:SetTitle(title)
	T.UI.filters[findex]:SetFGColor(fgcolor, tab_tbl.colorsource)
	T.UI.filters[findex]:SetBGColor(bgcolor, tab_tbl.colorsource)

	if (T.tab_index > 1) then
		exec_tab_func(tab_tbl.ChangeFilter, tab_tbl, findex, fpage)
	end
end

local function change_target(i)
	if (i == 0) then return end
	if (T.tab_index == 1) then
		if (T.tab_target_list[i]) then
			radar.SetRadarSelection(T.tab_target_list[i]["nodeID"], T.tab_target_list[i]["objectID"])
		end
	else
		local tab_tbl = T.tab_identifiers[T.tab_index]
		exec_tab_func(tab_tbl.ChangeTarget, tab_tbl, T.current_target_listindex, i)
	end
end

local function target_next()
	local num = #T.tab_target_list
	if (num == 0) then return end
	local i = T.current_target_listindex or 0
	i = (i + 1) % T.list_count
	if (i == 0) then i = T.list_count end
	if (i > num) then i = 1 end
	if (T.tab_target_list[i]["distance"] == 9999) then
		for j = i + 1, num do
			if (T.tab_target_list[j]["distance"] ~= 9999) then
				i = j
				break
			else
				i = 1
			end
		end
	end

	change_target(i)
end

local function target_prev()
	local num = #T.tab_target_list
	if (num == 0) then return end
	local i = T.current_target_listindex or 0
	i = (i - 1) % T.list_count
	if (i == 0) then i = T.list_count end
	if (i > num) then i = num end
	if (T.tab_target_list[i]["distance"] == 9999) then
		for j = i - 1, 1, -1 do
			if (T.tab_target_list[j]["distance"] ~= 9999) then
				i = j
				break
			else
				i = 0
			end
		end
	end
	change_target(i)
end

local function pin_target(_, data)
	data = data and tonumber(data[1])
	if (not T.tab_target_list[data]) then data = nil end

	if (T.tab_index == 1 and data or T.current_target_listindex) then
		local charID = T.tab_target_list[data or T.current_target_listindex]["charID"]
		if (T.pinlist[charID]) then
			T.pinlist[charID] = nil
		else
			T.pinlist[charID] = os.time()
		end
	end
end

local verb_color = {[true] = "\127FF0000", [false] = "\12700FF00"}

local verb = {
	{[true] = "only Hive", [false] = "all NPCs"},
	{[true] = "HIDING", [false] = "SHOWING"}, -- reversed logic because the variable is hide_turrets
	{[true] = "NPC ship types", [false] = "NPC names"},
	{[true] = "INVERTED", [false] = "NORMAL"},
}

local function toggle_hive_only()
	T.show_hive_only = not T.show_hive_only
	print(string.format("\127FFFFFFTargetList is now showing %s%s\127o on the NPC filter.", verb_color[T.show_hive_only], verb[1][T.show_hive_only]))
	T:SaveSettings()
end

local function toggle_friendly()
	T.hide_friendly = not T.hide_friendly
	print(string.format("\127FFFFFFTargetList is now %s%s\127o friendly targets.", verb_color[T.hide_friendly], verb[2][T.hide_friendly]))
	T:SaveSettings()
end

local function toggle_group()
	T.hide_group = not T.hide_group
	print(string.format("\127FFFFFFTargetList is now %s%s\127o group members.", verb_color[T.hide_group], verb[2][T.hide_group]))
	T:SaveSettings()
end

local function toggle_turrets()
	T.hide_turrets = not T.hide_turrets
	print(string.format("\127FFFFFFTargetList is now %s%s\127o turrets.", verb_color[T.hide_turrets], verb[2][T.hide_turrets]))
	T:SaveSettings()
end

local function toggle_npc_ship_types()
	T.show_npc_ship_types = not T.show_npc_ship_types
	print(string.format("\127FFFFFFTargetList is now showing %s%s\127o.", verb_color[not T.show_npc_ship_types], verb[3][T.show_npc_ship_types]))
	T:SaveSettings()
end

local function invert_faction_filter()
	T.invert_faction_filter = not T.invert_faction_filter
	print(string.format("\127FFFFFFTargetList faction filter is now %s%s\127o.", verb_color[T.invert_faction_filter], verb[4][T.invert_faction_filter]))
	iup.GetParent(T.UI.filters)[1].bgcolor = "255 0 0 "..(T.invert_faction_filter and 128 or 0)
	T:SaveSettings()
end

local function set_key_binds()
	for i = 1, T.list_count < 10 and T.list_count or 10 do
		gkinterface.GKProcessCommand("bind "..(i % 10).." target"..i)
	end
	gkinterface.GKProcessCommand("bind - cyclefilter")
	gkinterface.GKProcessCommand("bind = cyclefilterpage")
end

for i = 1, T.list_count < 10 and T.list_count or 10 do
	RegisterUserCommand("target"..i, change_target, i)
end

RegisterUserCommand("cyclefilter", cycle_filter)
RegisterUserCommand("cyclefilterpage", cycle_filter_page)
RegisterUserCommand("tabnext", tab_next)
RegisterUserCommand("tabprev", tab_prev)
RegisterUserCommand("invertfactionfilter", invert_faction_filter)
RegisterUserCommand("targetnext", target_next)
RegisterUserCommand("targetprev", target_prev)
RegisterUserCommand("pintarget", pin_target)
RegisterUserCommand("togglehiveonly", toggle_hive_only)
RegisterUserCommand("togglefriendly", toggle_friendly)
RegisterUserCommand("togglegroup", toggle_group)
RegisterUserCommand("toggleturrets", toggle_turrets)
RegisterUserCommand("togglenpcshiptypes", toggle_npc_ship_types)
RegisterUserCommand("settargetbinds", set_key_binds)
RegisterUserCommand("targetlist", cmd_processor)

return change_target
